#!/usr/bin/env python
# encoding: utf-8
"""
metahoard.common.model

Created by Keoni on 2009-03-31.
Copyright (c) 2009 University of Wisconsin SSEC. All rights reserved.

Credit where due: Modeled after some of the fine source code from byCycle.org

http://postgis.refractions.net/documentation/manual-1.3/ch02.html

on tahiti:

export PATH=$HOME/opt/postgis-1.3.3/bin:$PATH
export PGPORT=8330

# create query user for read-only access, password "anonymous"
createuser -P -S -D -R query
# create ingest user for read-write purposes, password in rayg's notes "Instrument Databases metahoard setup"
createuser -P -S -D -R ingest

# create tables and roles
dropdb metahoard
export DB_URI=postgres://localhost/metahoard
python -m metahoard.common.model -d ${DB_URI} create 
python -m metahoard.common.model -d ${DB_URI} build
python -m metahoard.common.model -d ${DB_URI} grant query select
python -m metahoard.common.model -d ${DB_URI} grant ingest all

# create indexes

"""
from __future__ import with_statement

_SVN_IDENT = "$Id: model.py 23 2009-04-03 16:04:20Z rayg $"

import sys, os, time
import logging
import uuid, re

from subprocess import check_call as sh

import psycopg2
from sqlalchemy import *
from sqlalchemy.orm import relation, sessionmaker, scoped_session
from sqlalchemy.types import Integer, String, CHAR, Integer
from sqlalchemy.ext.declarative import declarative_base

LOG = logging.getLogger( __name__ )

from metahoard.common.config import *
from metahoard.common.gis_types import *

from keoni.map.parse import parse_coords, format_coords, COORDINATE_FORMAT_DOC   # see larch.ssec.wisc.edu
from keoni.time.epoch import datetime_from_string


################################################################
##
## orm schema
##
################################################################

metadata = MetaData()
DeclarativeBase = declarative_base(metadata=metadata) #metadata=metadata, cls=Entity)
CASCADE_ARGS = 'all, delete-orphan'
SCHEMA = 'public' # postgres puts tables in public schema by default

def _tablename(name, schema=SCHEMA):
    return name
    # return schema + '.' + name

class Granule(DeclarativeBase):
    """a downloadable product
    represented as a collection of Fields
    a collection of Image
    a collection of Resources
    and a collection of Tags
    """
    __tablename__ = _tablename('granules')
    
    id = Column(Integer, primary_key=True)
    uu = Column(Uuid, unique=True, nullable=False)
    name = Column(String, unique=True, nullable=False)
    instrument = Column(String)
    perimeter = Column(Geometry)
    start_time = Column(DateTime)
    end_time = Column(DateTime)
    contact = Column(String)
    valid = Column(Boolean)
    
    fovs = relation('Fov', backref='granule', order_by='Fov.cross_track_index', cascade=CASCADE_ARGS)
    resources = relation('Resource', backref='granule', order_by='Resource.incept', cascade=CASCADE_ARGS)
    images = relation('Images', backref='granule', order_by='Image.start', cascade=CASCADE_ARGS)
    tags = relation('Tag', backref='granule', cascade=CASCADE_ARGS)

    def __repr__(self):
        return u'<granule %s>' % repr(self.name)


class Fov(DeclarativeBase):
    """
    a single field of view
    having one footprint latitude/longitude, 
    a field of regard (cross-track) number and a detector number
    """
    __tablename__ = _tablename('fields')
    id = Column(Integer, primary_key=True)
    granule_id = Column(Integer, ForeignKey('granules.id'))

    time = Column(DateTime, nullable=False)
    center = Column(Point, nullable=False)    
    cross_track_index = Column(Integer) # 0..
    along_track_index = Column(Integer) # 0..
    detector_index = Column(Integer)    # 0..
    
    surface_type = Column(SmallInteger)
    clear_percent = Column(SmallInteger)
    cloudy_percent = Column(SmallInteger)
    surface_kelvin = Column(SmallInteger)
    twometer_kelvin = Column(SmallInteger)
    surface_hpa = Column(SmallInteger)
    day_time = Column(SmallInteger)
    
    def __repr__(self):
        return u'<fov c%da%dd%d of %s>' % ( self.cross_track_index, 
                                            self.along_track_index, 
                                            self.detector_index, 
                                            repr(self.granule)
                                            )
    

class Image(DeclarativeBase):
    """
    an image of one or more bands, 
    optionally with geolocation or projection information
    as well as Resources to JPG/GeoTIFF/PNG resources"
    """
    __tablename__ = _tablename('images')
    id = Column(Integer, primary_key=True)
    granule_id = Column(Integer, ForeignKey('granules.id'))
    
    name = Column(UnicodeText)
    description = Column(UnicodeText)
    bands_json = Column(UnicodeText)
    start_time = Column(DateTime, nullable=False)
    end_time = Column(DateTime, nullable=False)
    perimeter = Column(Geometry) # null implies in-file
    projection = Column(UnicodeText) # null implies in-file
    uu = Column(Uuid) # for Resources

    def __repr__(self):
        return u'<image "%s" of %s>' % (self.name, repr(self.granule))

    
class Resource(DeclarativeBase):
    "an access location for a granule or Image"
    __tablename__ = _tablename('resources')
    id = Column(Integer, primary_key=True)
    granule_id = Column(Integer, ForeignKey('granules.id'))
    represents = Column(Uuid)  # granule, resource or tag uuid
    
    uri = Column(UnicodeText, unique=True, nullable=False)
    mimetype = Column(String)
    name = Column(UnicodeText, nullable=False)
    description = Column(UnicodeText)
    incept = Column(DateTime, nullable=False)
    expiry = Column(DateTime)
    verified = Column(DateTime)
    
    def __repr__(self):
        return u'<replica "%s" at "%s" of %s>' % (self.name, self.uri, repr(self.granule))
        

class Tag(DeclarativeBase):
    """a (key,name) => json-value triple associated with a granule
    allows multiple instances of a name (e.g. 'comment')
    or multiple names for a single logical key
    """
    __tablename__ = _tablename('tags')
    # id = Column(Integer, primary_key=True)
    granule_id = Column(Integer, ForeignKey('granules.id'))
    owner = Column(Uuid) # for Resources associated with the tag

    key = Column(Uuid, primary_key=True)
    name = Column(Unicode(64), primary_key=True)
    value_json = Column(UnicodeText)
    
    def __repr__(self):
        return u'<tag "%s":"%s" of %s>' % (self.name, self.key, repr(self.granule))


################################################################
##
## module front end functions - modeled after byCycle.org
##
################################################################

engine = None
connection = None
cursor = None
_session_class = None
session = None

def init(uri = None, **connection_args):
    # based on code from Wyatt Baldwin, byCycle.org <wyatt@bycycle.org>
    global engine, connection, cursor, session, _session_class
    engine = create_engine(uri or DEFAULT_DB_URI)
    connection = engine.raw_connection()
    cursor = connection.cursor()
    _session_class = sessionmaker(autoflush=True, autocommit=True, bind=engine)
    session = scoped_session(_session_class)

def connect(uri):
    metadata.bind = uri
    
def drop_all():
    metadata.drop_all()
    
def create_all():
    metadata.create_all()

def drop_schema(schema=SCHEMA, cascade=False):
    # based on code from Wyatt Baldwin, byCycle.org <wyatt@bycycle.org>
    cascade_clause = ' CASCADE' if cascade else ''
    Q = 'DROP SCHEMA %s%s' % (schema, cascade_clause)
    try:
        execute(Q)
    except psycopg2.ProgrammingError:
        rollback()  # important!
    else:
        commit()

def create_schema(schema=SCHEMA):
    # based on code from Wyatt Baldwin, byCycle.org <wyatt@bycycle.org>
    Q = 'CREATE SCHEMA %s' % schema
    try:
        execute(Q)
    except psycopg2.ProgrammingError:
        rollback()  # important!
    else:
        commit()

def vacuum():
    connection.set_isolation_level(0)
    cursor.execute("VACUUM FULL ANALYZE")
    connection.set_isolation_level(2)

def commit():
    connection.commit()

def rollback():
    connection.rollback()

def execute(sql):
    cursor.execute(sql)


################################################################
##
## postgres utility functions
##
################################################################

def _bash(script):
    fp = os.tmpfile()
    fp.write(script)
    fp.seek(0)
    sh(['/bin/bash'], stdin = fp)

def _psql(db,script,*args):
    fp = os.tmpfile()
    fp.write(script)
    fp.seek(0)
    sh(['psql', '-d', db] + list(args), stdin = fp)

def pg_createdb(db='metahoard'):
    if not os.path.exists('lwpostgis.sql'):
        LOG.error('need to be in CWD with lwpostgis.sql!')
        return
    script="""# GIS database initialize
    createdb %(db)s --encoding='UTF8' || exit 9
    createlang plpgsql %(db)s || exit 9
    psql -d %(db)s -f lwpostgis.sql  || exit 9
    psql -d %(db)s -f spatial_ref_sys.sql || exit 9
    """ % locals()
    _bash(script)

def grant_user(db='metahoard', name='query', perm='select', schema=SCHEMA, *args):
    "grant user permissions to database tables"
    script = """
      GRANT USAGE ON SCHEMA %(schema)s TO %(name)s;
      GRANT %(perm)s ON fields TO "%(name)s";
      GRANT %(perm)s ON fields_id_seq TO "%(name)s";
      GRANT %(perm)s ON granules TO "%(name)s";
      GRANT %(perm)s ON granules_id_seq TO "%(name)s";
      GRANT %(perm)s ON geometry_columns TO "%(name)s";
      GRANT %(perm)s ON tags TO "%(name)s";
      GRANT %(perm)s ON images TO "%(name)s";
      GRANT %(perm)s ON images_id_seq TO "%(name)s";
      GRANT %(perm)s ON resources TO "%(name)s";
      GRANT %(perm)s ON resources_id_seq TO "%(name)s";
      GRANT %(perm)s ON spatial_ref_sys TO "%(name)s";
    """ % locals()
    if connection: 
        execute(script)
        commit()
    else:
        LOG.info('using psql to grant permissions')
        _psql(db,script,*args)

def create_indices(db='metahoard'):
    "create indices for tables"
    script = """
        CREATE INDEX tags_owner_idx ON tags(owner);
        CREATE INDEX granules_name_idx ON granules(name);
        CREATE INDEX granules_uu_idx ON granules(uu);
        CREATE INDEX granules_start_idx ON granules(start_time);
        CREATE INDEX granules_end_idx ON granules(end_time);
    """
    # CREATE INDEX fields_center_gidx ON fields USING GIST(center);
    # CREATE INDEX granules_geom_gidx ON granules USING GIST(geom);
    # CREATE INDEX images_shape_gidx ON images USING GIST(geom);
    if connection: 
        execute(script)
        commit()
    else:
        LOG.info('using psql to change indices')
        _psql(db,script,*args)

def add_geometry_column(table, srid, geom_type, name='geom', schema=SCHEMA, create_index=True):
    """Add a PostGIS geometry column to ``table``.

    ``table``
        SQLAlchemy ``Table``

    ``srid``
        Spatial reference ID

    ``geom_type``
        POINT, LINESTRING, etc

    ``name``
        Name to give the new geometry column

    """
    # Add geometry columns after tables are created and add gist INDEXes them
    drop_col = 'ALTER TABLE "%s"."%s" DROP COLUMN %s'
    add_geom_col = "SELECT AddGeometryColumn('%s', '%s', '%s', %s, '%s', 2)"
    create_gist_index = ('CREATE INDEX "%s_%s_gist"'
                         'ON "%s"."%s"'
                         'USING GIST ("%s" gist_geometry_ops)')
    geom_type = geom_type.upper()
    try:
        execute(drop_col % (schema, table, name))
    except psycopg2.ProgrammingError:
        rollback()  # important!
    execute(add_geom_col % (schema, table, name, srid, geom_type))
    execute(create_gist_index % (table, name, schema, table, name))
    commit()
    
    
def setup_gis_columns():
    "replace columns with GIS datatypes and build indexes"
    add_geometry_column( 'granules', LONLAT_SRID, 'POLYGON', 'perimeter' )
    add_geometry_column( 'images', LONLAT_SRID, 'POLYGON', 'perimeter' )
    add_geometry_column( 'fields', LONLAT_SRID, 'POINT', 'center' )






################################################################
##
## tests and module main
##
################################################################


    
def test1(db_uri=None, *args, **kwargs):
    metadata.bind = db_uri
    metadata.create_all()
    g = model.Granule(uu=model.uuid.uuid1(), name='test_granule', instrument='test', valid=False)
    session.save(g)
    session.add(g)
    session.commit()




def main():
    import optparse
    usage = """
%prog [options] 
run "%prog help" to list commands
"""
    parser = optparse.OptionParser(usage)
    parser.add_option('-t', '--test', dest="self_test",
                    action="store_true", default=False, help="run unit tests")            
    parser.add_option('-q', '--quiet', dest="quiet",
                    action="store_true", default=False, help="only error output")
    parser.add_option('-v', '--verbose', dest="verbose",
                    action="store_true", default=False, help="enable more informational output")   
    parser.add_option('-w', '--debug', dest="debug",
                    action="store_true", default=False, help="enable debug output")   
    parser.add_option('-d', '--database', dest="db_uri",
                    default=DEFAULT_DB_URI, help="URI of PostGIS database")   
    options, args = parser.parse_args()
    if options.self_test:
        import doctest
        doctest.testmod()
        sys.exit(2)

    lvl = logging.WARNING
    if options.debug: lvl = logging.DEBUG
    elif options.verbose: lvl = logging.INFO
    elif options.quiet: lvl = logging.ERROR
    logging.basicConfig(level = lvl)

    db = options.db_uri.split('/')[-1]
    
    commands = {}
    prior = None
    prior = dict(locals())
    
    def create(*args):
        """create empty database on local postgis using createdb
        """
        LOG.info("creating database %s" % db)
        pg_createdb(db=db,*args)
        
    def build(*args):
        """build database tables 
        """
        LOG.info('initializing connection to server')
        init(options.db_uri)
        LOG.info("connecting to database at URI %s" % options.db_uri)
        connect(options.db_uri)
        LOG.info("creating schema")
        create_schema()
        LOG.info("creating tables")
        create_all()
        LOG.info("replacing columns with GIS types")
        setup_gis_columns()
        LOG.info("creating additional indexes")
        create_indices()
        
    def grant(*args):
        """grant {user} {permissions} e.g. grant query select using psql
        """
        LOG.info("granting permissions for tables")
        grant_user(db,*args)     
                
    def help(command=None):
        """print command help or list of commands
        e.g. help help
        """
        if command is None: 
            # print first line of docstring
            for cmd in commands:
                ds = commands[cmd].__doc__.split('\n')[0]
                print "%-16s %s" % (cmd,ds)
            # print ', '.join(commands.keys())
        else:
            # import pydoc
            print commands[command].__doc__
            # print globals()[command].__doc__
            
    def test():
        "run tests"
        test1()
        
    commands.update(dict(x for x in locals().items() if x[0] not in prior))    
    
    if not args: 
        parser.print_help()
        help()
    else:
        locals()[args[0]](*args[1:])

    # FIXME(args,**vars(options))

    return 0


if __name__=='__main__':
    sys.exit(main())
